def test_imports_ok() -> None:
    # This test verifies that cli at least imports properly on
    # in all supported envs and python versions
    #
    # Will be replaced with some proper e2e tests
    import neuro_flow.cli  # noqa
